import Text.ParserCombinators.ReadP
import Data.Char (isSpace, isLetter, isDigit, digitToInt)
import Data.List (union, delete)
import Data.Maybe (fromJust)

-- auxiliary function on tuples
mapPair :: (a -> b) -> (c -> d) -> (a, c) -> (b, d)
mapPair f g (x, y) = (f x, g y)

mapFst :: (a -> b) -> (a, c) -> (b, c)
mapFst f = mapPair f id

mapSnd :: (c -> d) -> (a, c) -> (a, d)
mapSnd = mapPair id


-- state monad
data State s a = State (s -> (a, s))

runState :: State s a -> s -> (a, s)
runState (State m) = m

instance Monad (State s) where
  return x = State (\s -> (x, s))
  m >>= k = State (\s -> let (x, s') = (runState m) s 
                         in (runState (k x)) s')

getState :: State s s
getState = State (\s -> (s, s))

putState :: s -> State s () 
putState t = State (\_ -> ((), t))

evalState :: State s a -> s -> a
evalState m = fst . runState m


-- parser combinators
greedyMany :: ReadP a -> ReadP [a]
greedyMany p = greedyMany1 p <++ return []

greedyMany1 :: ReadP a -> ReadP [a]
greedyMany1 p = do
  x <- p
  xs <- greedyMany p
  return (x : xs)

parseParen p = do
  skipSpaces
  char '('
  skipSpaces
  x <- p
  skipSpaces
  char ')'
  skipSpaces
  return x

-- characters and escaping
specials :: String 
specials = "%.()\\0123456789-,+*"
isDelimiter c = c `elem` specials || isSpace c

reserved :: [String]
reserved = ["if", "True", "False", "then", "else", "fst", "snd", "iszero"]

escapeChar :: Char -> String
escapeChar c
  | isDelimiter c = '\\' : c : []
  | otherwise     = [c]

escape :: String -> String
escape v
  | v `elem` reserved = '\\' : v
  | otherwise         = concatMap escapeChar v


-- Mini-Haskell terms
type Variable = String

data BinOp = Plus | Times

evalBinOp :: BinOp -> Integer -> Integer -> Integer
evalBinOp Plus  i j = i + j
evalBinOp Times i j = i * j

data Term
  = Id Variable
  | Ap Term Term
  | Lam Variable Term
  | Lit Integer
  | Tuple Term Term
  | Fst Term
  | Snd Term
  | IsZero Term
  | TRUE
  | FALSE
  | IfThenElse Term Term Term
  | BinOp Term BinOp Term

foldTerm
    :: (Variable -> a)        -- Id
    -> (a -> a -> a)          -- Ap
    -> (Variable -> a -> a)   -- Lam
    -> (Integer -> a)         -- Lit
    -> (a -> a -> a)          -- Tuple
    -> (a -> a)               -- Fst
    -> (a -> a)               -- Snd
    -> (a -> a)               -- IsZero
    -> a                      -- TRUE
    -> a                      -- FALSE
    -> (a -> a -> a -> a)     -- IfThenElse
    -> (a -> BinOp -> a -> a) -- BinOp
    -> Term
    -> a
foldTerm id ap lam lit tuple fst snd iszero true false ifthenelse binop = go
  where
    go (Id v) = id v
    go (Ap s t) = ap (go s) (go t)
    go (Lam v t) = lam v (go t)
    go (Lit i) = lit i
    go (Tuple s t) = tuple (go s) (go t)
    go (Fst t) = fst (go t)
    go (Snd t) = snd (go t)
    go (IsZero t) = iszero (go t)
    go TRUE = true
    go FALSE = false
    go (IfThenElse t1 t2 t3) = ifthenelse (go t1) (go t2) (go t3)
    go (BinOp t1 bop t2) = binop (go t1) bop (go t2)

-- free variables of a term
free :: Term -> [String]
free = foldTerm (:[]) union delete (const []) union id id id [] [] union3 union13
  where
    union3 a b c = a `union` b `union` c
    union13 a _ b = a `union` b

{-
Pretty-printing syntax, also used for parsing:
  Lam x t is written as %x. t
  Ap t t' is written as t t'
  Lit i is written as i in decimal notation, possibly with a preceding - sign, but not a + sign.
  Tuple t t' is written as (t, t')
  Fst t is written as fst t
  Snd t is written as snd t
  IsZero t is written as iszero t
  TRUE is written as True
  FALSE is written as False
  Binary operators (+) and (*) have to be written prefix with parentheses.
  Predefined functions fst, snd, iszero, and binary operators cannot be partially applied.
  Id x is written as x with special characters escaped by \.
  Special characters are all white space and those listed in special.
-}

renderTerm :: Int -> Term -> ShowS
renderTerm _ (Id v) = (escape v++)
renderTerm n (Ap t1 t2) =
  showParen (n > 9) (renderTerm 9 t1 . (' ':) . renderTerm 10 t2)
renderTerm n (Lit i) = showsPrec n i
renderTerm n (Tuple t1 t2) = showParen True (renderTerm 0 t1 . (", "++) . renderTerm 0 t2)
renderTerm n (Fst t) =
  showParen (n > 9) (("fst "++) . renderTerm 10 t)
renderTerm n (Snd t) =
  showParen (n > 9) (("snd "++) . renderTerm 10 t)
renderTerm n (IsZero t) =
  showParen (n > 9) (("iszero "++) . renderTerm 10 t)
renderTerm n TRUE = ("True"++)
renderTerm n FALSE = ("False"++)
renderTerm n (IfThenElse t1 t2 t3) =
  showParen (n > 9) (("if "++) . renderTerm 8 t1 . (" then "++) . renderTerm 8 t2 . (" else "++) . renderTerm 8 t3)
renderTerm n (BinOp t1 bop t2) = 
  showParen (n > 9) (("("++) . (show bop++) . (") "++) . renderTerm 10 t1 . (" "++) . renderTerm 10 t2)
renderTerm n t =
  showParen (n > 0) (('%':) . (unwords vars++) . (". "++) . renderTerm 0 body)
  where
    (vars, body) = collectLams t
    collectLams (Lam x t) = mapFst (escape x:) (collectLams t)
    collectLams t = ([], t)

parseTerm :: ReadP Term
parseTerm = between skipSpaces skipSpaces term
  where
    idLetter = (do
       char '\\'
       c <- get
       return (True, c)
     ) <++ (do
       c <- satisfy (not . isDelimiter)
       return (False, c)
     )
    identifier = between skipSpaces skipSpaces $ do
      escapeVar <- greedyMany1 idLetter
      let (escapes, var) = unzip escapeVar
      if var `notElem` reserved || or escapes then return var else pfail

    integer = do
      sign <- option 1 (char '-' >> return (-1))
      digits <- greedyMany1 (satisfy isDigit >>= return . digitToInt)
      skipSpaces
      return (Lit (sign * foldl (\rest digit -> rest * 10 + toInteger digit) 0 digits))

    id = identifier >>= return . Id

    tuple = do
      char '('
      skipSpaces
      t1 <- term
      skipSpaces
      char ','
      skipSpaces
      t2 <- term
      skipSpaces
      char ')'
      return (Tuple t1 t2)

    fst = string "fst" >> skipSpaces >> atom >>= return . Fst
    snd = string "snd" >> skipSpaces >> atom >>= return . Snd
    iszero = string "iszero" >> skipSpaces >> atom >>= return . IsZero

    true = string "True" >> skipSpaces >> return TRUE
    false = string "False" >> skipSpaces >> return FALSE
    ifthenelse = do
      string "if"
      skipSpaces
      cnd <- apAtom
      skipSpaces
      string "then"
      skipSpaces
      thn <- apAtom
      skipSpaces
      string "else"
      skipSpaces
      els <- term
      return (IfThenElse cnd thn els)

    binop' = (string "(+)" >> return Plus) +++ (string "(*)" >> return Times)
    binop = do
      bop <- binop'
      skipSpaces
      left <- atom
      skipSpaces
      right <- atom
      return (BinOp left bop right)

    term = lam +++ apAtom +++ ifthenelse +++ binop
    apAtom = do
      t <- atom
      ts <- many atom
      return (foldl Ap t ts)
    atom = parseParen term +++ id +++ integer +++ tuple +++ fst +++ snd +++ iszero +++ true +++ false

    lam = do
      char '%'
      ids <- greedyMany1 identifier
      char '.'
      skipSpaces
      t <- term
      return (foldr Lam t ids)

instance Show Term where
  showsPrec = renderTerm

instance Read Term where
  readsPrec _ = readP_to_S parseTerm

instance Show BinOp where
  showsPrec _ Plus = ("+"++)
  showsPrec _ Times = ("*"++)


-- evaluation
subst           -- substitutes
  :: Variable   -- variable with
  -> Term       -- term in
  -> Term       -- term 
  -> Term
subst v s (Id x)                = if x == v then s else Id x
subst v s (Ap t1 t2)            = Ap (subst v s t1) (subst v s t2)
subst _ _ (Lit i)               = Lit i
subst v s (Tuple t1 t2)         = Tuple (subst v s t1) (subst v s t2)
subst v s (Fst t)               = Fst (subst v s t)
subst v s (Snd t)               = Snd (subst v s t)
subst v s (IsZero t)            = IsZero (subst v s t)
subst v s TRUE                  = TRUE
subst v s FALSE                 = FALSE
subst v s (IfThenElse t1 t2 t3) = IfThenElse (subst v s t1) (subst v s t2) (subst v s t3)
subst v s (BinOp t1 bop t2)     = BinOp (subst v s t1) bop (subst v s t2)
subst v s (Lam x t)
  | x == v                = Lam x t
  | not (x `elem` free s) = Lam x (subst v s t)
  | otherwise             = Lam z (subst v s (subst x (Id z) t))
  where
    z = fresh (free t ++ free s)
    fresh vs = foldr max "" vs ++ "'"

eager :: Term -> Term
eager (Id x)       = Id x
eager (Lit i)      = Lit i
eager (Tuple t t') = Tuple (eager t) (eager t')
eager (Fst t)      = let r = eager t in case r of { Tuple t' _ -> t'; _ -> Fst r }
eager (Snd t)      = let r = eager t in case r of { Tuple _ t' -> t'; _ -> Snd r }
eager (IsZero t)   = let r = eager t in case r of { Lit n -> if n == 0 then TRUE else FALSE; _ -> IsZero r }
eager TRUE         = TRUE
eager FALSE        = FALSE
eager (IfThenElse t1 t2 t3) =
  let r1 = eager t1 in case r1 of { TRUE -> eager t2; FALSE -> eager t3; _ -> IfThenElse r1 t2 t3 }
eager (BinOp t1 bop t2) =
  let r1 = eager t1; r2 = eager t2
  in case (r1, r2) of { (Lit i1, Lit i2) -> Lit (evalBinOp bop i1 i2); _ -> BinOp r1 bop r2 }
eager (Ap t t')    =
  case r of
    Lam v body -> eager (subst v r' body)
    _          -> Ap r r'
  where
    r  = eager t
    r' = eager t'
eager (Lam x t)    = Lam x (eager t)

lazy :: Term -> Term
lazy (Id x)        = Id x
lazy (Lit i)       = Lit i
lazy (Tuple t t')  = Tuple t t'
lazy (Fst t)       = let r = lazy t in case r of { Tuple t' _ -> lazy t'; _ -> Fst r }
lazy (Snd t)       = let r = lazy t in case r of { Tuple _ t' -> lazy t'; _ -> Snd r }
lazy (IsZero t)    = let r = lazy t in case r of { Lit n -> if n == 0 then TRUE else FALSE; _ -> IsZero r }
lazy TRUE          = TRUE
lazy FALSE         = FALSE
lazy (IfThenElse t1 t2 t3) =
  let r1 = lazy t1 in case r1 of { TRUE -> lazy t2; FALSE -> lazy t3; _ -> IfThenElse r1 t2 t3 }
lazy (BinOp t1 bop t2) =
  let r1 = lazy t1; r2 = lazy t2
  in case (r1, r2) of { (Lit i1, Lit i2) -> Lit (evalBinOp bop i1 i2); _ -> BinOp r1 bop r2 }
lazy (Ap t t')     =
  case r of
    Lam v body -> lazy (subst v t' body)
    _          -> Ap r t'
  where
    r = lazy t
lazy (Lam v t)     = Lam v t -- no evaluation under a lambda abstraction



-- Mini-Haskell types
type TyVar = Integer
data Type = TyVar TyVar | TyFun Type Type | TyInt | TyBool | TyTuple Type Type
  deriving (Eq)

foldType
  :: (TyVar -> a)    -- TyVar
  -> (a -> a -> a)   -- TyFun
  -> a               -- TyInt
  -> a               -- TyBool
  -> (a -> a -> a)   -- TyTuple
  -> Type
  -> a
foldType tyvar tyfun tyint tybool tytuple = go
  where
    go (TyVar v)         = tyvar v
    go (TyFun ty1 ty2)   = tyfun (go ty1) (go ty2)
    go TyInt             = tyint
    go TyBool            = tybool
    go (TyTuple ty1 ty2) = tytuple (go ty1) (go ty2)

-- pretty printing
renderType :: Int -> Type -> ShowS
renderType n (TyVar i) = showsPrec n i
renderType n (TyFun ty1 ty2) = showParen (n > 9) (renderType 10 ty1 . (" -> "++) . renderType 9 ty2)
renderType _ TyInt = ("Int"++)
renderType _ TyBool = ("Bool"++)
renderType _ (TyTuple ty1 ty2) =
  showParen True (renderType 0 ty1 . (", "++) . renderType 0 ty2)

instance Show Type where
  showsPrec = renderType

-- type variables
tyVars :: Type -> [TyVar]   -- list of type variables in a type
tyVars = foldType (:[]) union [] [] union

tySubst  -- replace type variable with type in type
  :: TyVar  -- substitution variable
  -> Type   -- substituted type
  -> Type   -- type to substitute in
  -> Type
tySubst v s = foldType substVar TyFun TyInt TyBool TyTuple
  where
    substVar x = if v == x then s else TyVar x

-- increments the state and returns the new value
nextTyVar :: State Integer Integer
nextTyVar = do
  v <- getState
  let v' = v + 1
  putState v'
  return v'

-- type inference

type Context = [(Variable, Type)]   -- associative list
type Constraint = (Type, Type)      -- equality constraint on types
type Substitution = [(TyVar, Type)] -- associative list

-- collect all constraints from typing rules
collectConstraints :: Context -> Term -> Type -> [Constraint]
collectConstraints gamma t ty = evalState (collect gamma t ty) v
  where
    v = max (foldr max 0 (map (maxVar . snd) gamma)) (maxVar ty)
    maxVar = foldr max 0 . tyVars

    collect :: Context -> Term -> Type -> State Integer [Constraint]
    collect gamma (Id v) ty =
      case lookup v gamma of
        Nothing -> error ("Free variable " ++ show v)
        Just tyV -> return [(tyV, ty)]
    collect gamma (Lit i) ty = return [(TyInt, ty)]
    collect gamma (Tuple fst snd) ty = do
      n <- nextTyVar
      n' <- nextTyVar
      cs1 <- collect gamma fst (TyVar n)
      cs2 <- collect gamma snd (TyVar n')
      return ((ty, TyTuple (TyVar n) (TyVar n')) : cs1 ++ cs2)
    collect gamma (Fst t) ty = do
      n <- nextTyVar
      collect gamma t (TyTuple ty (TyVar n))
    collect gamma (Snd t) ty = do
      n <- nextTyVar
      collect gamma t (TyTuple (TyVar n) ty)
    collect gamma (IsZero t) ty = do
      cs <- collect gamma t TyInt
      return ((ty, TyBool) : cs)
    collect gamma TRUE ty = return [(TyBool, ty)]
    collect gamma FALSE ty = return [(TyBool, ty)]
    collect gamma (IfThenElse t1 t2 t3) ty = do
      csC <- collect gamma t1 TyBool
      csT <- collect gamma t2 ty
      csF <- collect gamma t3 ty
      return (csC ++ csT ++ csF)
    collect gamma (BinOp t1 bop t2) ty = do
      csL <- collect gamma t1 TyInt
      csR <- collect gamma t2 TyInt
      return ((TyInt, ty) : csL ++ csR)
    collect gamma (Lam v t) ty = do
      n <- nextTyVar
      n' <- nextTyVar
      cs <- collect ((v, TyVar n) : gamma) t (TyVar n')
      return ((ty, TyFun (TyVar n) (TyVar n')) : cs)
    collect gamma (Ap t t') ty = do
      n <- nextTyVar
      cs1 <- collect gamma t (TyFun (TyVar n) ty)
      cs2 <- collect gamma t' (TyVar n)
      return (cs1 ++ cs2)

-- solve list of constraits by unification, returns most general unifier
solveConstraints
  :: [Constraint]                -- List of constraints to solve
  -> Substitution                -- Accumulator for substitution.
  -> Substitution
solveConstraints []       sigma = sigma
solveConstraints (c : cs) sigma =
  case c of
    (TyVar v, ty) ->
      if ty == TyVar v then
        solveConstraints cs sigma  -- trivial constraint
      else if v `elem` tyVars ty then
        error ("occurs check for " ++ show v ++ " = " ++ show ty)
      else
        {- substitute ty for v in remaining contraints,
           record assignment,
           and substitute ty for v in previous assignments -}
        solveConstraints
          (substConstraints v ty cs)
          ((v, ty) : substSubstitution v ty sigma)
    (t,             TyVar v) -> solveConstraints ((TyVar v, t) : cs) sigma
    (TyBool,        TyBool)          -> solveConstraints cs sigma
    (TyInt,         TyInt)           -> solveConstraints cs sigma
    (TyFun t1 t2,   TyFun t1' t2')   -> solveConstraints ((t1, t1') : (t2, t2') : cs) sigma
    (TyTuple t1 t2, TyTuple t1' t2') -> solveConstraints ((t1, t1'):(t2, t2'):cs) sigma
    (ty,            ty')             ->
      error ("unification failed for " ++ show ty ++ " = " ++ show ty')
  where
    substConstraints v ty = map (mapPair (tySubst v ty) (tySubst v ty))
    substSubstitution v ty = map (mapSnd (tySubst v ty))

inferType :: Term -> Type
inferType t =
  let constraints = collectConstraints [] t (TyVar 0)
      sigma = solveConstraints constraints []
  in fromJust (lookup 0 sigma)


-- Tests applied to a list
readall input = map read input :: [Term]
inferall = map (inferType . read) 
lazyall  = map (lazy . read)
eagerall = map (eager . read)

-- readall and inferall work
inferTests = [
  "%x . True" ,
  "%x y z. (x y) z" ,
  "%f. fst (f,f)" ,
  "%x. (x 1 True, x 0)" ,
  "(%x. %y. (y (iszero (y x)))) True" ,
  "%x y . if y x then (fst x) else (snd (snd x))" ,
  "%x . x (%y . x y)" ,
  "%x . (%y . y) x" ,
  "%x y z. (y x, x z)" ,
  "%x y. x (%z . y)" ,
  "(%x. %y. x y) (%z. y z)" ,
  "(%f. %x. x f) (%r. r) ((%y. %z. y z) (%a. a))"
  ]
            
-- readall works, but inferall returns an error for each of them
failToInferTests = [
  "iszero (fst ((+) 3 5))" ,
  "%x . x (%y . x y)" ,
  "(%x . x (%y . x y)) (%x . (%y . y) x)" ,
  "(%x. %y. x y) (%z. y z)" ,
  "%x. (fst x, iszero x)" ,
  "%f. f f"
  ]

-- eagerall and lazyall work ; but inferall does not
evalTests = [
  "(%x. %y. x y) (%z. y z)" ,
  "(%f. %x. x f) (%r. r) ((%y. %z. y z) (%a. a))" ,
  "(%x . x (%y . x y)) (%x . (%y . y) x)" ,
  "(%x. %y. (y (iszero (y x)))) True"
  ]

-- all work
inferAndEvalTests = [
  "(if (iszero ((*)2 0)) then (fst (2,3)) else (snd (2,3)))" ,
  "((%f x . (if (iszero x) then (f 2) else (f 3))) ((%x y . ((+) y x)) 2) 5)" ,
  "(%z. z) (%x. %y. x)" ,
  "%x y z . x z (y z)" ,
  "(%z. z) (%x y. x)" ,
  "%x. iszero ((fst x) (2, True))" ,
  "%x. %y. (iszero (x y), x 3)" ,
  "(%n . if iszero n then 1 else ((*) 2 n)) ((%x . ((+) x 2)) (fst (2,True)))" ,
  "%p . (snd p) (fst p)" 
  ]
 